#!/usr/bin/python

"""
Usage:
    # L3 clear commands
    quagga clear bgp [vrf <text>] (<bgppeer>|all) [soft in|soft out]
    quagga clear bgp [vrf <text>] prefix <ipv4/prefixlen>
    quagga clear ospf interface [<interface>]

    # L3 show commands
    quagga show bgp [vrf <text>] [summary]
    quagga show bgp [vrf <text>] (<ipv4>|<ipv4/prefixlen>) [bestpath|multipath] [json]
    quagga show bgp (<ipv6>|<ipv6/prefixlen>) [bestpath|multipath] [json]
    quagga show bgp ipv4 unicast [<ipv4>|<ipv4/prefixlen>|summary] [json]
    quagga show bgp ipv6 unicast [<ipv6>|<ipv6/prefixlen>|summary] [json]
    quagga show bgp [vrf <text>] neighbor [<bgppeer>]
    quagga show bgp [vrf <text>] update-groups
    quagga show ospf [json]
    quagga show ospf database (asbr-summary|external|network|nssa-external|opaque-area|opaque-as|opaque-link|router|summary) [<ipv4>|adv-router <ipv4-adv-router>|self-originate]
    quagga show ospf database [max-age|self-originate]
    quagga show ospf interface [<interface>] [json]
    quagga show ospf neighbor [all|<interface>|<ipv4>|detail] [json]
    quagga show ospf route
    quagga show ospf6
    quagga show ospf6 database [adv-router <ipv4>|linkstate-id <ipv4-id>|self-originated] [detail|dump|internal]
    quagga show ospf6 interface [<interface>]
    quagga show ospf6 linkstate [detail|router <ipv4>]
    quagga show ospf6 neighbor [detail|drchoice]
    quagga show ospf6 redistribute
    quagga show ospf6 route [<ipv6>|<ipv6/prefixlen>|detail|summary]
    quagga show ospf6 route (intra-area|inter-area|external-1|external-2) [detail]
    quagga show ospf6 route <ipv6/prefixlen> longer [detail]
    quagga show ospf6 route <ipv6/prefixlen> match [detail]
    quagga show ospf6 spf tree
    quagga show route [vrf <text>] [<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>|bgp|connected|kernel|ospf|ospf6|ripng|static|summary]
    quagga show route [vrf <text>] (<ipv4/prefixlen>|<ipv6/prefixlen>) longer-prefixes

    # L3 debug commands
    quagga add       debug bgp bestpath <ip/prefixlen>
    quagga del       debug bgp bestpath [<ip/prefixlen>]
    quagga add       debug bgp keepalives (<ip>|<interface>)
    quagga del       debug bgp keepalives [<ip>|<interface>]
    quagga add       debug bgp neighbor-events (<ip>|<interface>)
    quagga del       debug bgp neighbor-events [<ip>|<interface>]
    quagga add       debug bgp updates prefix <ip/prefixlen>
    quagga del       debug bgp updates prefix [<ip/prefixlen>]
    quagga add       debug bgp zebra prefix <ip/prefixlen>
    quagga del       debug bgp zebra prefix [<ip/prefixlen>]
    quagga (add|del) debug ospf ism [status|events|timers]
    quagga (add|del) debug ospf lsa [generate|flooding|install|refresh]
    quagga (add|del) debug ospf nsm [status|events|timers]
    quagga (add|del) debug ospf packet (hello|dd|ls-request|ls-update|ls-ack|all) [send|recv|detail]
    quagga (add|del) debug ospf zebra [interface|redistribute]
    quagga (add|del) debug ospf6 (abr|asbr|flooding|interface)
    quagga (add|del) debug ospf6 border-routers [area-id <ipv4>|router-id <ipv4>]
    quagga (add|del) debug ospf6 lsa (router|network|inter-prefix|inter-router|as-external|link|intra-prefix|unknown) [originate|examine|flooding]
    quagga (add|del) debug ospf6 message (unknown|hello|dbdesc|lsreq|lsupdate|lsack|all) [send|recv]
    quagga (add|del) debug ospf6 neighbor [state|event]
    quagga (add|del) debug ospf6 route (table|intra-area|inter-area|memory)
    quagga (add|del) debug ospf6 spf (database|process|time)
    quagga (add|del) debug ospf6 zebra [send|recv]


    # L3 global configuration
    quagga add       router-id <ipv4>
    quagga del       router-id [<ipv4>]
    quagga (add|del) ip protocol (bgp|ospf|static) route-map <route-map>
    quagga (add|del) ip route <ipv4/prefixlen> (<ipv4>|<interface>|Null0|blackhole|reject) [<1-255>] [vrf <text>]
    quagga (add|del) ipv6 route <ipv6/prefixlen> (<ipv6>|<interface>) [blackhole|reject] [<1-255>] [vrf <text>]
    quagga (add|del) ipv6 route <ipv6/prefixlen> <ipv6> <interface> [blackhole|reject] [<1-255>] [vrf <text>]
    quagga (add|del) log file <text> [alerts|critical|debugging|emergencies|errors|informational|notifications|warnings]
    quagga (add|del) log timestamp precision <0-6>
    quagga (add|del) ptm-enable
    quagga (add|del) interface <interface> [no] ptm-enable
    quagga add       interface <interface> ipv6 nd ra-interval (<1-1800>|msec <70-1800000>)
    quagga del       interface <interface> ipv6 nd ra-interval [<1-1800>]
    quagga del       interface <interface> ipv6 nd ra-interval msec [<70-1800000>]
    quagga add       interface <interface> ipv6 nd ra-lifetime <0-9000>
    quagga del       interface <interface> ipv6 nd ra-lifetime [<0-9000>]
    quagga (add|del) interface <interface> ipv6 nd suppress-ra

    # L3 BGP 'global' configuration
    quagga (add|del) router bgp <1-4294967295> [vrf <text>]
    quagga (add|del) bgp listen limit <1-5000>
    quagga (add|del) bgp listen range (<ipv4/prefixlen>|<ipv6/prefixlen>) peer-group <peergroup>
    quagga (add|del) bgp default ipv4-unicast

    quagga (add|del) bgp [vrf <text>] autonomous-system <1-4294967295>
    quagga (add|del) bgp [vrf <text>] always-compare-med
    quagga (add|del) bgp [vrf <text>] bestpath as-path multipath-relax [as-set|no-as-set]
    quagga (add|del) bgp [vrf <text>] bestpath compare-routerid
    quagga (add|del) bgp [vrf <text>] bestpath med missing-as-worst
    quagga add       bgp [vrf <text>] neighbor <bgppeer> advertisement-interval <0-600>
    quagga del       bgp [vrf <text>] neighbor <bgppeer> advertisement-interval [<0-600>]
    quagga (add|del) bgp [vrf <text>] neighbor <bgppeer> bfd [<2-255> <50-60000> <50-60000>]
    quagga (add|del) bgp [vrf <text>] neighbor <bgppeer> bfd [multihop]
    quagga (add|del) bgp [vrf <text>] neighbor <bgppeer> capability extended-nexthop
    quagga add       bgp [vrf <text>] neighbor <bgppeer> description <wildcard>
    quagga del       bgp [vrf <text>] neighbor <bgppeer> description [<wildcard>]
    quagga (add|del) bgp [vrf <text>] neighbor <bgppeer> ebgp-multihop [<1-255>]
    quagga add       bgp [vrf <text>] neighbor <bgppeer> password <text-password>
    quagga del       bgp [vrf <text>] neighbor <bgppeer> password [<text-password>]
    quagga (add|del) bgp [vrf <text>] neighbor (<ip>|<bgppeer>|<interface>) peer-group <peergroup>
    quagga (add|del) bgp [vrf <text>] neighbor (<ip>|<bgppeer>|<interface>) remote-as (<1-4294967295>|external|internal)
    quagga (add|del) bgp [vrf <text>] neighbor <bgppeer> shutdown
    quagga add       bgp [vrf <text>] neighbor <bgppeer> timers <0-65535> <3-65535>
    quagga del       bgp [vrf <text>] neighbor <bgppeer> timers [<0-65535>] [<3-65535>]
    quagga add       bgp [vrf <text>] neighbor <bgppeer> timers connect <1-65535>
    quagga del       bgp [vrf <text>] neighbor <bgppeer> timers connect [<1-65535>]
    quagga (add|del) bgp [vrf <text>] neighbor <bgppeer> ttl-security hops <1-254>
    quagga add       bgp [vrf <text>] neighbor <bgppeer> update-source (<ipv4>|<ipv6>|<interface-source>)
    quagga del       bgp [vrf <text>] neighbor <bgppeer> update-source [<ipv4>|<ipv6>|<interface-source>]
    quagga (add|del) bgp [vrf <text>] neighbor <interface> interface v6only
    quagga (add|del) bgp [vrf <text>] neighbor <interface> interface peer-group <peergroup>
    quagga (add|del) bgp [vrf <text>] neighbor <interface> interface remote-as (<1-4294967295>|external|internal)
    quagga (add|del) bgp [vrf <text>] neighbor <text-peer-group> peer-group
    quagga add       bgp [vrf <text>] router-id <ipv4>
    quagga del       bgp [vrf <text>] router-id [<ipv4>]
    quagga add       bgp [vrf <text>] update-delay <0-3600> [<1-3600>]
    quagga del       bgp [vrf <text>] update-delay <0-3600> <1-3600>
    quagga del       bgp [vrf <text>] update-delay [<0-3600>]

    # L3 BGP per address-family configuration
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast] aggregate-address <ipv4/prefixlen> [as-set] [summary-only]
    quagga (add|del) bgp [vrf <text>] ipv6 unicast aggregate-address <ipv6/prefixlen> [summary-only]
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast] network <ipv4/prefixlen> [route-map <text-route-map>]
    quagga (add|del) bgp [vrf <text>] ipv6 unicast network <ipv6/prefixlen> [route-map <text-route-map>]
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast] redistribute (connected|isis|kernel|ospf|pim|rip|static|table) [route-map <route-map>]
    quagga (add|del) bgp [vrf <text>] ipv6 unicast redistribute (connected|isis|kernel|ospf6|ripng|static|table) [route-map <route-map>]
    quagga add       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] maximum-paths [ibgp] <1-64>
    quagga del       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] maximum-paths [ibgp] [<1-64>]
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> activate
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> addpath-tx-all-paths
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> addpath-tx-bestpath-per-AS
    quagga add       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> allowas-in <1-10>
    quagga del       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> allowas-in [<1-10>]
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> default-originate
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> default-originate route-map <route-map>
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> filter-list <as-path-access-list> (in|out)
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> next-hop-self [force]
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast] neighbor <bgppeer> prefix-list <prefix-list-v4> (in|out)
    quagga (add|del) bgp [vrf <text>] ipv6 unicast neighbor <bgppeer> prefix-list <prefix-list-v6> (in|out)
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> remove-private-AS [all] [replace-AS]
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> route-map <route-map> (in|out)
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> route-reflector-client
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> send-community [both|extended|standard]
    quagga (add|del) bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] neighbor <bgppeer> soft-reconfiguration inbound
    quagga add       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] table-map <route-map>
    quagga del       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] table-map [<route-map>]

    # L3 OSPF configuration
    quagga (add|del) router ospf
    quagga (add|del) ospf area (<ipv4>|<0-4294967295>) stub [no-summary]
    quagga (add|del) ospf log-adjacency-changes [detail]
    quagga (add|del) ospf network <ipv4/prefixlen> area (<ipv4>|<0-4294967295>)
    quagga del       ospf network <ipv4/prefixlen>
    quagga (add|del) ospf passive-interface <interface>
    quagga (add|del) ospf redistribute (bgp|connected|isis|kernel|rip|static|table) [route-map <route-map>]
    quagga add       ospf router-id <ipv4>
    quagga del       ospf router-id [<ipv4>]
    quagga (add|del) ospf timers lsa min-arrival <0-600000>
    quagga add       ospf timers throttle spf <0-600000> <0-600000> <0-600000>
    quagga del       ospf timers throttle spf [<0-600000> <0-600000> <0-600000>]
    quagga add       interface <interface> ip ospf area (<ipv4>|<0-4294967295>)
    quagga del       interface <interface> ip ospf area [<ipv4>|<0-4294967295>]
    quagga (add|del) interface <interface> ip ospf bfd [<2-255> <50-60000> <50-60000>]

    quagga add       interface <interface> ip ospf dead-interval <1-65535>
    quagga del       interface <interface> ip ospf dead-interval [<1-65535>]
    quagga add       interface <interface> ip ospf hello-interval <1-65535>
    quagga del       interface <interface> ip ospf hello-interval [<1-65535>]
    quagga add       interface <interface> ip ospf network (broadcast|non-broadcast|point-to-multipoint|point-to-point)
    quagga del       interface <interface> ip ospf network [broadcast|non-broadcast|point-to-multipoint|point-to-point]

    # L3 OSPFv3 configuration
    quagga (add|del) router ospf6
    quagga (add|del) ospf6 area (<ipv4>|<0-4294967295>) stub [no-summary]
    quagga (add|del) ospf6 auto-cost reference-bandwidth <1-4294967>
    quagga (add|del) ospf6 interface <interface> area <ipv4>
    quagga (add|del) ospf6 log-adjacency-changes [detail]
    quagga (add|del) ospf6 redistribute (bgp|connected|isis|kernel|ripng|static|table) [route-map <route-map>]
    quagga add       ospf6 router-id <ipv4>
    quagga del       ospf6 router-id [<ipv4>]
    quagga (add|del) ospf6 stub-router administrative
    quagga (add|del) ospf6 timers lsa min-arrival <0-600000>
    quagga add       ospf6 timers throttle spf <0-600000> <0-600000> <0-600000>
    quagga del       ospf6 timers throttle spf [<0-600000> <0-600000> <0-600000>]
    quagga (add|del) interface <interface> ipv6 ospf6 bfd [<2-255> <50-60000> <50-60000>]
    quagga add       interface <interface> ipv6 ospf6 cost <1-65535>
    quagga del       interface <interface> ipv6 ospf6 cost [<1-65535>]
    quagga add       interface <interface> ipv6 ospf6 dead-interval <1-65535>
    quagga del       interface <interface> ipv6 ospf6 dead-interval [<1-65535>]
    quagga add       interface <interface> ipv6 ospf6 hello-interval <1-65535>
    quagga del       interface <interface> ipv6 ospf6 hello-interval [<1-65535>]
    quagga add       interface <interface> ipv6 ospf6 ifmtu <1-65535>
    quagga del       interface <interface> ipv6 ospf6 ifmtu [<1-65535>]
    quagga add       interface <interface> ipv6 ospf6 instance-id <0-255>
    quagga del       interface <interface> ipv6 ospf6 instance-id [<0-255>]
    quagga (add|del) interface <interface> ipv6 ospf6 mtu-ignore
    quagga add       interface <interface> ipv6 ospf6 network (broadcast|point-to-point)
    quagga del       interface <interface> ipv6 ospf6 network [broadcast|point-to-point]
    quagga (add|del) interface <interface> ipv6 ospf6 passive
    quagga add       interface <interface> ipv6 ospf6 priority <0-255>
    quagga del       interface <interface> ipv6 ospf6 priority [<0-255>]
    quagga add       interface <interface> ipv6 ospf6 retransmit-interval <1-65535>
    quagga del       interface <interface> ipv6 ospf6 retransmit-interval [<1-65535>]
    quagga add       interface <interface> ipv6 ospf6 transmit-delay <1-3600>
    quagga del       interface <interface> ipv6 ospf6 transmit-delay [<1-3600>]

    # as-path lists
    quagga (add|del) ip as-path access-list <text> (deny|permit) <wildcard>
    quagga show ip as-path-access-list [<as-path-access-list>]

    # community-list
    quagga (add|del) ip community-list standard <text> (deny|permit) (local-AS|no-advertise|no-export|internet|<community>)
    quagga (add|del) ip community-list expanded <text> (deny|permit) <regex>
    quagga (add|del) ip extcommunity-list standard <text> (deny|permit) (rt|soo) <extcommunity>
    quagga (add|del) ip extcommunity-list expanded <text> (deny|permit) <regex>

    quagga show ip community-list <community-list>
    quagga show ip extcommunity-list <extcommunity-list>

    # prefix-list
    quagga (add|del) ip prefix-list <text> description <wildcard>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) (<ipv4/prefixlen>|any)
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> ge <0-32>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> ge <0-32> le <1-32>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> le <0-32>
    quagga (add|del) ip prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> le <0-32> ge <1-32>
    quagga (add|del) ipv6 prefix-list <text> description <wildcard>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) (<ipv6/prefixlen>|any)
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> ge <0-128>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> ge <0-128> le <1-128>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> le <0-128>
    quagga (add|del) ipv6 prefix-list <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> le <0-128> ge <1-128>
    quagga clear ip prefix-list [<prefix-list-v4>|<prefix-list-v4> <ipv4/prefixlen>]
    quagga clear ipv6 prefix-list [<prefix-list-v6>|<prefix-list-v6> <ipv6/prefixlen>]
    quagga show ip prefix-list [<prefix-list-v4>]
    quagga show ip prefix-list <prefix-list-v4> seq <1-4294967295>
    quagga show ip prefix-list <prefix-list-v4> <ipv4/prefixlen> [longer|first-match]
    quagga show ip prefix-list summary [<prefix-list-v4>]
    quagga show ip prefix-list detail [<prefix-list-v4>]
    quagga show ipv6 prefix-list [<prefix-list-v6>]
    quagga show ipv6 prefix-list <prefix-list-v6> seq <1-4294967295>
    quagga show ipv6 prefix-list <prefix-list-v6> <ipv6/prefixlen> [longer|first-match]
    quagga show ipv6 prefix-list summary [<prefix-list-v6>]
    quagga show ipv6 prefix-list detail [<prefix-list-v6>]

    # route-map
    quagga (add|del) route-map <text> (deny|permit) <1-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> call <route-map>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> continue <0-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> description <wildcard>
    quagga del route-map <text> permit <1-65535> on-match (goto <0-65535>|next)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match as-path <as-path-access-list>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match community <community-list> [exact-match]
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match extcommunity <extcommunity-list>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match interface <interface>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ip (address|next-hop) prefix-len <0-32>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ip (address|next-hop|route-source) prefix-list <prefix-list-v4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ipv6 address prefix-list <prefix-list-v6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match ipv6 next-hop <ipv6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match local-preference <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match metric <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match origin (egp|igp|incomplete)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match peer (<ipv4>|<ipv6>|local)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match probability <0-100>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match source-protocol (bgp|connected|isis|kernel|ospf|ospf6|rip|ripng|static|system)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> match tag <0-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set aggregator as <1-4294967295> <ipv4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set as-path exclude <1-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set as-path prepend (<1-4294967295>|last-as <1-10>)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set atomic-aggregate
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set comm-list <community-list> delete
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set community <community> [additive]
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set community none
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set extcommunity (rt|soo) <extcommunity>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set forwarding-address <ipv6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set ip next-hop (<ipv4>|peer-address|unchanged)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set ipv6 next-hop (global|local) <ipv6>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set ipv6 next-hop peer-address
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set local-preference <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set metric <0-4294967295>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set metric-type (type-1|type-2)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set origin (egp|igp|incomplete)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set originator-id <ipv4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set src (<ipv4>|<ipv6>)
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set tag <0-65535>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set vpnv4 next-hop <ipv4>
    quagga (add|del) route-map <text> (deny|permit) <1-65535> set weight <0-4294967295>
    quagga show route-map [<route-map>]


Options:

    <as-path-access-list>              : An AS_PATH access-list
    <bgppeer>                          : A BGP peer or neighbor
    <community-list>                   : A BGP community-list
    <community>                        : XX:YY where XX and YY are numbers in the range 0-65535
    <extcommunity-list>                : A BGP extended community-list
    <extcommunity>                     : ASN:nn_or_IP-address:nn VPN extended community
    <interface-source>                 : Source interface
    <ipv4-adv-router>                  : OSPF advertising router
    <ipv4-id>                          : OSPFv3 advertising router
    <ipv4-mcast-group>                 : IPv4 multicast group
    <ipv4-source>                      : IPv4 source address
    <peergroup>                        : A BGP peer-group name
    <prefix-list-v4>                   : An IPv4 prefix-list name
    <prefix-list-v6>                   : An IPv6 prefix-list name
    <regex>                            : A regular expression
    <route-map>                        : A route-map name
    <text-password>                    : A password
    <text-peer-group>                  : A BGP peer-group name
    <text-route-map>                   : A route-map name
    Null0                              : Null interface
    abr                                : Area Border Router
    activate                           : Enable the Address Family for this Neighbor
    additive                           : Add to the existing list of communities
    addpath-tx-all-paths               : Use addpath to advertise all paths to a neighbor
    addpath-tx-bestpath-per-AS         : Use addpath to advertise the bestpath per each neighboring AS
    administrative                     : Advertise inability to be a transit router
    adv-router                         : Advertising Router link states
    advertisement-interval             : Minimum interval between sending BGP routing updates
    aggregate-address                  : Configure BGP aggregate entries
    aggregator                         : BGP aggregator attribute
    alerts                             : Immediate action needed
    allowas-in                         : Accept as-path with my AS present in it
    always-compare-med                 : Allow comparing MED from different neighbors
    any                                : Any IP prefix
    area                               : An OSPF area
    area-id                            : An OSPF area ID
    as                                 : Autonomous system
    as-external                        : AS-External LSA's
    as-override                        : Override ASNs in outbound updates if aspath equals remote-as
    as-path                            : AS_PATH attribute
    as-path-access-list                : An AS_PATH access-list
    as-set                             : Generate an AS_SET
    asbr                               : Autonomous system border router
    asbr-summary                       : ASBR summary link states
    assert                             : PIM interface assert
    assert-internal                    : PIM interface internal assert state
    assert-metric                      : PIM interface assert metric
    assert-winner-metric               : PIM interface assert winner metric
    atomic-aggregate                   : BGP atomic aggregate attribute
    attribute-unchanged                : BGP attribute is propagated unchanged to this neighbor
    auto-cost                          : Calculate OSPF interface cost according to bandwidth
    autonomous-system                  : AS number
    bestpath                           : BGP bestpath
    bfd                                : Enables BFD support
    bgp                                : Border Gateway Protocol
    blackhole                          : Silently discard pkts when matched
    border-routers                     : OSPF ABR's and ASBR's
    both                               : Both standard and extended communities
    broadcast                          : An OSPF broadcast multi-access network
    call                               : Jump to another Route-Map after match+set
    capability                         : Advertise capability to the peer
    client-to-client                   : Configure client to client route reflection
    cluster-id                         : Configure Route-Reflector Cluster-id
    comm-list                          : A BGP community list
    community                          : BGP community attribute
    community-list                     : Add a community list entry
    compare-routerid                   : Compare router-id for identical EBGP paths
    confed                             : Compare path lengths including confederation sets & sequences in selecting a route
    confederation                      : Confederation parameters
    connect                            : BGP connect timer
    connected                          : Connected routes (directly attached subnet or host)
    continue                           : Exit policy on matches
    cost                               : Interface cost
    count                              : Route and packet count data
    critical                           : Critical conditions
    dampening                          : Enable route-flap dampening
    database                           : OSPF database
    dbdesc                             : Database description message
    dd                                 : OSPF Database Description
    dead-interval                      : Interval after which a neighbor is declared dead
    debug                              : Enable debugs
    debugging                          : Debugging messages
    default                            : Configure defaults
    default-originate                  : Originate default route to this neighbor
    deny                               : Deny in routing policy
    description                        : Description
    designated-router                  : PIM interface designated router
    deterministic-med                  : Pick the best-MED path among paths advertised from the neighboring AS
    disable-connected-check            : One-hop away EBGP peer using loopback address
    disable-ebgp-connected-route-check : Disable checking if nexthop is connected on ebgp sessions
    distribute-list                    : Filter updates to/from this neighbor
    dont-capability-negotiate          : Do not perform capability negotiation
    drchoice                           : Designated Router choices
    drpriority                         : PIM Designated Router Election Priority
    dump                               : Dump LSAs
    ebgp-multihop                      : Allow EBGP neighbors not on directly connected networks
    egp                                : Exterior Gateway Protocol
    emergencies                        : System is unusable
    enforce-first-as                   : Enforce the first AS for EBGP routes
    enforce-multihop                   : Enforce EBGP neighbors perform multihop
    errors                             : Error conditions
    event                              : Event information
    events                             : Event information
    exact-match                        : Do exact matching of communities
    examine                            : Examine
    exclude                            : Exclude from the as-path
    expanded                           : An expanded community-list
    extcommunity                       : An extended community
    extcommunity-list                  : An extended community list
    extended                           : Send Extended Community attributes
    extended-nexthop                   : Advertise extended next-hop capability to the peer
    external                           : An external BGP peer, external OSPF LSA, etc
    external-1                         : OSPF Type-1 External routes
    external-2                         : OSPF Type-2 External routes
    fast-external-failover             : Immediately reset session if a link to a directly connected external peer goes down
    file                               : Logging to file
    filter-list                        : Apply an as-path access-list
    first-match                        : First matched prefix
    flooding                           : LSA Flooding
    force                              : Set the next hop to self for reflected routes
    forwarding-address                 : Forwarding Address
    ge                                 : Greater than or equal to
    generate                           : LSA Generation
    global                             : IPv6 global address
    goto                               : Goto Clause number
    graceful-restart                   : Graceful restart capability parameters
    groups                             : IGMP groups information
    hello                              : PIM interface hello information
    hello-interval                     : Time between HELLO packets
    hops                               : TTL hop count
    ibgp                               : An internal BGP peer
    ifmtu                              : Interface MTU
    igmp                               : Internet Group Management Protocol
    ignore                             : Ignore as-path length in selecting a route
    igp                                : Interior Gateway Protocol
    import-table                       : Import routes from non-main kernel table
    in                                 : Apply inbound
    inbound                            : Apply inbound
    incomplete                         : Unknown origin
    informational                      : Informational messages
    install                            : LSA Install/Delete
    instance-id                        : Instance identifier
    inter-area                         : OSPF Inter-Area routes
    inter-prefix                       : Inter-Prefix LSA's
    inter-router                       : Inter-Router LSA's
    interface                          : An interface such as swp1, swp2, etc
    internal                           : An internal BGP peer
    internet                           : BGP community, matches all routes
    intra-area                         : OSPF Intra-Area routes
    intra-prefix                       : Intra-Prefix LSA's
    ip                                 : Internet Protocol version 4
    ipv4                               : Internet Protocol version 4
    ipv4-unicast                       : IPv4 Unicast address family
    ipv6                               : Internet Protocol version 6
    isis                               : Intermediate System to Intermediate System (IS-IS)
    ism                                : Interface State Machine
    join                               : PIM interface join information
    jp-override-interval               : PIM interface J/P override interval
    keepalives                         : BGP keepalives
    kernel                             : Kernel routes (not installed via the zebra RIB)
    lan-prune-delay                    : PIM neighbors LAN prune delay parameters
    last-as                            : Use the peer's AS-number
    le                                 : Less than or equal to
    limit                              : Maximum number of BGP Dynamic Neighbors that can be created
    linkstate                          : Display linkstate routing table
    linkstate-id                       : Search by Link state ID
    listen                             : Configure BGP Dynamic Neighbors
    local                              : IPv6 local address
    local-AS                           : BGP community, restrict routes to our confederations sub-AS
    local-as                           : Specify a local-as number
    local-membership                   : PIM interface local-membership
    local-preference                   : BGP local-preference attribute
    log                                : Logging control
    log-adjacency-changes              : Log neighbor up/down and reset reason
    log-neighbor-changes               : Log neighbor up/down and reset reason
    longer                             : Show route matching the specified Network/Mask pair only
    longer-prefixes                    : Show route matching the specified Network/Mask pair only
    ls-ack                             : Link State Acknowledgement
    ls-request                         : Link State Request
    ls-update                          : Link State Update
    lsa                                : Link State Advertisement
    lsack                              : Link State Acknowledgement
    lsreq                              : Link State Request
    lsupdate                           : Link State Update
    match                              : Match values from routing table
    max-age                            : LSAs in MaxAge list
    max-med                            : Advertise routes with max-med
    maximum-paths                      : Forward packets over multiple paths
    maximum-prefix                     : Maximum number of prefix accept from this peer
    med                                : MED attribute
    memory                             : Memory statistics
    message                            : OSPFv3 message
    metric                             : Metric value for destination routing protocol
    metric-type                        : Type of metric
    min-arrival                        : Minimum delay in receiving new version of a LSA
    missing-as-worst                   : Treat missing MED as the least preferred one
    mroute                             : Configure static unicast route into MRIB for multicast RPF lookup
    msdp                               : Multicast Source Discovery Protocol
    msec                               : Millisecond
    mtu-ignore                         : Ignore MTU
    multicast                          : Multicast options
    multicast-routing                  : Enable IP multicast forwarding
    multihop                           : Neighbor is not directly connected
    multipath                          : BGP paths flagged as multipath
    multipath-relax                    : Allow load sharing across routes that have different AS paths (but same length)
    nd                                 : Neighbor Discovery
    neighbor                           : A BGP, OSPF, PIM, etc neighbor
    neighbor-events                    : Neighbor state transition events
    network                            : An ipv4 or ipv6 network
    next                               : Next route-map clause
    next-hop                           : Next-hop address of route
    next-hop-self                      : Disable the next hop calculation for this neighbor
    nht                                : Next Hop Tracking
    no-advertise                       : BGP community, do not advertise to anyone
    no-as-set                          : Do not generate an AS_SET
    no-export                          : BGP community, do not advertise to eBGP peer
    no-summary                         : Do not inject inter-area routes into stub
    non-broadcast                      : An OSPF NBMA network
    none                               : No community attribute
    notifications                      : Normal but significant conditions
    nsm                                : Neighbor State Machine
    nssa                               : OSPF nssa information
    nssa-external                      : NSSA external link state
    on-match                           : Exit policy on matches
    opaque-area                        : Link area Opaque-LSA
    opaque-as                          : Link AS Opaque-LSA
    opaque-link                        : Link local Opaque-LSA
    origin                             : BGP origin attribute
    originate                          : Originate
    originator-id                      : BGP ORIGINATOR_ID attribute
    ospf                               : Open Shortest Path First (OSPFv2)
    ospf6                              : Open Shortest Path First (OSPFv3)
    out                                : Apply outbound
    override-capability                : Override capability negotiation result
    packet                             : Packet
    packets                            : Packets
    parameters                         : IGMP parameters information
    passive                            : Do not send open messages to this neighbor
    passive-interface                  : Suppress routing updates on an interface
    password                           : Set a password
    peer                               : Match peer address
    peer-address                       : Use peer address (for BGP only)
    peer-group                         : Member of the peer-group
    permit                             : Permit in routing policy
    pim                                : Protocol Independent Multicast
    point-to-multipoint                : An OSPF point-to-multipoint network
    point-to-point                     : An OSPF point-to-point network
    port                               : Neighbor's TCP port
    precision                          : Set the timestamp precision
    prefix                             : Clear bestpath and re-advertise
    prefix-len                         : Match prefix length of ip address
    prefix-list                        : Filter updates to/from this neighbor
    prepend                            : Prepend to the as-path
    priority                           : priority
    probability                        : Match portion of routes defined by percentage value
    process                            : Debug Detailed SPF Process
    protocol                           : Filter routing info exchanged between zebra and protocol
    ptm-enable                         : Enable neighbor check with specified topology
    query-interval                     : IGMP host query interval
    query-max-response-time            : IGMP max query response value (seconds)
    ra-interval                        : Router Advertisement interval
    ra-lifetime                        : Router lifetime
    range                              : Add a listening range for Dynamic Neighbors
    recv                               : Packet received
    redistribute                       : Redistribute information from another routing protocol
    reference-bandwidth                : Use reference bandwidth method to assign OSPF cost
    refresh                            : LSA Refresh
    reject                             : Emit an ICMP unreachable when matched
    remote-as                          : Specify a BGP neighbor
    remove-private-AS                  : Remove private ASNs in outbound updates
    replace-AS                         : Replace private ASNs with our ASN in outbound updates
    retransmit-interval                : Time between retransmitting lost link state advertisements
    rip                                : Routing Information Protocol (RIP)
    ripng                              : Routing Information Protocol Next Generation
    route                              : Static routes
    route-map                          : Route-map
    route-reflector                    : Allow modifications made by out route-map
    route-reflector-client             : Configure a neighbor as Route Reflector client
    route-server-client                : Configure a neighbor as Route Server client
    route-source                       : Match advertising source address of route
    router                             : L3 routing protocols or OSPF LSA type
    router-id                          : Router identifier
    rp                                 : Rendevous Point
    rp-info                            : PIM RP information
    rpf                                : PIM cached source rpf information
    rt                                 : Route Target extended community
    secondary                          : PIM neighbor addresses
    self-originate                     : Self-originated link states
    self-originated                    : Self-originated link states
    send                               : Packet sent
    send-community                     : Send Community attribute to this neighbor
    seq                                : Sequence number
    set                                : Set values in destination routing protocol
    shutdown                           : Administratively shut down this neighbor
    soft                               : Soft reconfig inbound and outbound updates
    soft-reconfiguration               : Per neighbor soft reconfiguration
    solo                               : Solo peer - part of its own update group
    soo                                : Site-of-Origin extended community
    source-protocol                    : Match protocol via which the route was learnt
    sources                            : IGMP sources information
    spf                                : OSPF SPF timers
    src                                : Source
    ssm                                : Source Specific Multicast
    standard                           : Standard BGP communities
    state                              : State information
    static                             : Statically configured routes
    strict-capability-match            : Strict capability negotiation match
    stub                               : Configure OSPF area as stub
    stub-router                        : Make router a stub router
    summary                            : Network summary link states
    summary-only                       : Filter more specific routes from updates
    suppress-ra                        : Suppress IPv6 Router Advertisement
    table                              : Non-main Kernel Routing Table
    table-map                          : BGP table to RIB route download filter
    tag                                : Set tag for this route
    throttle                           : Throttling adaptive timer
    time                               : Measure time taken by SPF Calculation
    timers                             : BGP per neighbor timers
    timestamp                          : Timestamp configuration
    trace                              : IGMP/PIM internal daemon activity
    transmit-delay                     : Transmit delay of this interface
    tree                               : Show SPF tree
    ttl-security                       : Specify the maximum number of hops to the BGP peer
    type-1                             : OSPF6 external type 1 metric
    type-2                             : OSPF6 external type 2 metric
    unchanged                          : Don't modify existing Next hop address
    unicast                            : Unicast Address Family Modifier
    unknown                            : Unknown LSA's
    unsuppress-map                     : Route-map to selectively unsuppress suppressed routes
    update-delay                       : Force initial delay for best-path and updates
    update-groups                      : BGP update groups
    update-source                      : Source of routing updates
    updates                            : BGP updates
    upstream                           : PIM upstream information
    upstream-join-desired              : PIM upstream join-desired
    upstream-rpf                       : PIM upstream source rpf
    v6only                             : Enable BGP with v6 link-local only
    vpnv4                              : Vpnv4 address family
    vrf                                : Virtual Routing and Forwarding
    warnings                           : Warning conditions
    weight                             : Set default weight for routes from this neighbor
    zebra                              : Zebra information

"""

# Syntax for access-list...we decided to leave these out since prefix-lists
# are a much better way for filtering routes.
'''
    # access-list
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> <ipv4-dest> <ipv4-dest-wildcard>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> any
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> host <ipv4-dest>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any <ipv4-dest> <ipv4-dest-wildcard>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any any
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any host <ipv4-dest>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> <ipv4-dest> <ipv4-dest-wildcard>
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> any
    quagga (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> host <ipv4-dest>
    quagga (add|del) access-list (<1-99>|<1300-1999>) (deny|permit) <ipv4-source> [<ipv4-source-wildcard>]
    quagga (add|del) access-list (<1-99>|<1300-1999>) (deny|permit) (any|host <ipv4-source>)
    quagga (add|del) access-list <text> (deny|permit) (any|<ipv4/prefixlen>|<ipv4/prefixlen> exact-match)
    quagga (add|del) ipv6 access-list <text> (deny|permit) (any|<ipv6/prefixlen>|<ipv6/prefixlen> exact-match)
    quagga show ip access-list [<1-2699>|<text>]
    quagga show ipv6 access-list [<text>]
'''

from collections import OrderedDict
from copy import deepcopy
from network_docopt import NetworkDocopt, get_network_docopt_info, sort_for_humans
from nclu import (print_diff,
    files_match,
    chown_chgroup,
    file_contains,
    ifname_is_valid,
    ifname_is_globish,
    ifname_is_glob,
    ifname_expand_glob,
    globify_interface_commands,
    Capturing,
    print_pending_add_del_commands)
from os import listdir
from re import search as re_search
from pprint import pformat, pprint
from shutil import copyfile
from subprocess import check_output, STDOUT, CalledProcessError
import json
import logging
import logging.handlers
import os
import re
import sys

# /usr/lib/quagga/quagga-reload.py
sys.path.append('/usr/lib/quagga/')
quagga_reload = __import__("quagga-reload")  # use __import__ for files with a hyphen in the name

log = logging.getLogger('netd')

# pim commands are added to __doc_ if /usr/lib/quagga/pimd exists
# This happens via build_doc_string()
pim_doc_string = """# L3 multicast debugs
    quagga (add|del) debug igmp [events|packets|trace]
    quagga (add|del) debug pim [events|packets|trace|zebra]

    # L3 multicast show commands
    quagga show igmp (join|sources)
    quagga show igmp groups [json]
    quagga show igmp interface [detail|<interface>] [json]
    quagga show mroute [count|json]
    quagga show msdp (peer|sa) [json]
    quagga show pim (assert|assert-internal|assert-metric|assert-winner-metric|secondary)
    quagga show pim (join|local-membership|rpf|rp-info|upstream|upstream-join-desired|upstream-rpf) [json]
    quagga show pim interface [detail|<interface>] [json]
    quagga show pim neighbor [detail|<ipv4>] [json]
    quagga show pim state [json]
    quagga show pim state <ipv4-source> [json]
    quagga show pim state <ipv4-source> <ipv4-group> [json]

    # L3 multicast configuration
    quagga (add|del) ip multicast-routing
    quagga add       ip pim rp <ipv4>
    quagga del       ip pim rp [<ipv4>]
    quagga (add|del) interface <interface> ip igmp
    quagga (add|del) interface <interface> ip igmp join <ipv4-mcast-group> <ipv4-source>
    quagga add       interface <interface> ip igmp query-interval <1-1800>
    quagga del       interface <interface> ip igmp query-interval [<1-1800>]
    quagga add       interface <interface> ip igmp query-max-response-time <1-25>
    quagga del       interface <interface> ip igmp query-max-response-time [<1-25>]
    quagga add       interface <interface> ip pim drpriority <1-4294967295>
    quagga del       interface <interface> ip pim drpriority [<1-4294967295>]
    quagga add       interface <interface> ip pim hello <1-180> <3-180>
    quagga del       interface <interface> ip pim hello [<1-180> <3-180>]
    quagga (add|del) interface <interface> ip pim (sm|ssm)
"""

etc_quagga_daemons = '/etc/quagga/daemons'
etc_quagga_quagga_conf = '/etc/quagga/Quagga.conf'
usr_lib_quagga_pimd = '/usr/lib/quagga/pimd'


def build_doc_string():
    global __doc__

    if os.path.exists(usr_lib_quagga_pimd):
        __doc__ = __doc__.replace('    # L3 OSPF configuration', pim_doc_string + '\n    # L3 OSPF configuration')


class ConfigWrapper(object):

    def __init__(self, show_linux_command, color_diffs):
        self.working_dir = '/var/run/nclu/quagga/'
        self.baseline_conf = os.path.join(self.working_dir, 'baseline.conf')
        self.pending_conf = os.path.join(self.working_dir, 'pending.conf')
        self.pending_baseline_conf = os.path.join(self.working_dir, 'pending_baseline.conf')
        self.baseline_daemons = os.path.join(self.working_dir, 'baseline.daemons')
        self.pending_daemons = os.path.join(self.working_dir, 'pending.daemons')
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs

        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir, mode=0600)

    def get_managed_files(self):
        files = []
        files.append(etc_quagga_daemons)
        files.append(etc_quagga_quagga_conf)
        return files

    def get_service_cmd_for_file(self, filename):
        cmds = []

        if filename == etc_quagga_daemons:
            cmds.append(('/bin/systemctl', 'reset-failed', 'quagga'))
            cmds.append(('/bin/systemctl', 'restart', 'quagga'))

        elif filename == etc_quagga_quagga_conf:
            cmds.append(('/bin/systemctl', 'reset-failed', 'quagga'))
            cmds.append(('/bin/systemctl', 'reload', 'quagga'))

        return cmds

    def init_guts(self, print_missing_file_errors=True):

        if not os.path.exists(etc_quagga_daemons) or os.path.getsize(etc_quagga_daemons) == 0:
            with open(etc_quagga_daemons, 'w') as fh:
                fh.write('zebra=yes\n')

        if not os.path.isfile(etc_quagga_quagga_conf):
            if print_missing_file_errors:
                print """ERROR: %s does not exist

net only supports quagga when it uses the single config file %s
To configure quagga to use %s please run the following:

sudo systemctl start quagga
sudo vtysh -c 'conf t' -c 'service integrated-vtysh-config'
sudo vtysh -c 'wr mem'
""" % (etc_quagga_quagga_conf, etc_quagga_quagga_conf, etc_quagga_quagga_conf)
            return False

        if not os.path.isfile(self.baseline_daemons):
            copyfile(etc_quagga_daemons, self.baseline_daemons)
            copyfile(self.baseline_daemons, self.pending_daemons)

        # Use baseline.conf to later determine if Quagga.conf has been modified
        if not os.path.isfile(self.baseline_conf):
            copyfile(etc_quagga_quagga_conf, self.baseline_conf)

        # In the beginning our staging file is a "marked" copy of Quagga.conf
        # "marked" means that each context ends with the word "end"
        if not os.path.isfile(self.pending_conf) or not os.path.isfile(self.pending_baseline_conf):
            try:
                output = check_output(['/usr/bin/vtysh',
                                       '-m',
                                       '-f',
                                       etc_quagga_quagga_conf],
                                       stderr=STDOUT).splitlines()
            except CalledProcessError as e:
                print "ERROR marking %s\n" % etc_quagga_quagga_conf
                print e.output
                return False
            else:
                with open(self.pending_conf, 'w') as fh:
                    for line in output:
                        if line != '!':
                            fh.write(line + '\n')

                with open(self.pending_baseline_conf, 'w') as fh:
                    for line in output:
                        if line != '!':
                            fh.write(line + '\n')

        # Verify quagga's config has not been modified
        if not files_match(etc_quagga_daemons, self.baseline_daemons):
            print "\nERROR: %s was modified by another user" % etc_quagga_daemons
            print_pending_add_del_commands()
            return False

        if not files_match(etc_quagga_quagga_conf, self.baseline_conf):
            print "\nERROR: %s was modified by another user" % etc_quagga_quagga_conf
            print_pending_add_del_commands()
            return False

        return True

    def look_for_line(self, target_line, exact_match):
        """
        Look for the target_line in pending_baseline_conf and pending_conf
        If we find the line, return it else return None
        """
        lines = []
        with open(self.pending_baseline_conf, 'r') as fh:
            for line in fh.readlines():
                line = line.strip()

                if exact_match:
                    if line == target_line:
                        lines.append(line)
                else:
                    if line.startswith(target_line):
                        lines.append(line)

        with open(self.pending_conf, 'r') as fh:
            for line in fh.readlines():
                line = line.strip()

                if exact_match:
                    if line == target_line:
                        lines.append(line)
                else:
                    if line.startswith(target_line):
                        lines.append(line)

        return lines

    def get_bgp_asn(self, vrf):
        asn = None

        if vrf:
            vrf = "vrf %s" % vrf

        for line in self.look_for_line('router bgp', False):
            # router bgp 100 vrf RED
            if vrf:
                if line.endswith(vrf):
                    return int(line.split()[2])
            else:
                if 'vrf' not in line:
                    return int(line.split()[2])

        return None

    def neighbor_interface(self, neighbor):
        """
        Return True if 'neighbor' is already configured as an interface neighbor
        """
        target_line = 'neighbor %s interface' % neighbor

        if self.look_for_line(target_line, False):
            return True
        else:
            return False

    def neighbor_remote_as(self, neighbor):
        """
        Return True if 'neighbor' already has a remote-as configured
        """

        # First see if the neighbor has remote-as configured directly on the neighbor
        neighbor_remote_as = 'neighbor %s remote-as' % neighbor

        if self.look_for_line(neighbor_remote_as, False):
            return True

        # If not, see if the peer is a member of a peer-group where the
        # peer-group has remote-as configured
        neighbor_peer_group = 'neighbor %s peer-group ' % neighbor
        neighbor_peer_group = self.look_for_line(neighbor_peer_group, False)

        if neighbor_peer_group:
            neighbor_peer_group = neighbor_peer_group[0]

            if neighbor_peer_group.endswith('peer-group'):
                peer_group = neighbor_peer_group.split()[1]
            else:
                peer_group = neighbor_peer_group.split()[3]
            peer_group_remote_as = 'neighbor %s remote-as' % peer_group

            if self.look_for_line(peer_group_remote_as, False):
                return True

        return False

    def edit_pending_file(self, target_context, target_subcontext, add_line, del_line, exact_match):

        if not self.init_guts():
            return False

        if target_subcontext:
            subcontext_display = "\n %s\n" % target_subcontext
        else:
            subcontext_display = ''

        if add_line:
            action = 'Adding'
            add_line = add_line.rstrip()
            line_display = "\n %s\n" % add_line
            found_del_line = True

        if del_line:
            action = 'Deleting'
            del_line = del_line.strip()
            line_display = "\n %s\n" % del_line
            found_del_line = False

        log.debug("%s '%s' %s %s" % (action, target_context, subcontext_display, line_display))

        # Stage an update to the daemons file.  Most of the time we won't need
        # to modify this file.
        daemons = OrderedDict()
        rewrite_daemons = False

        with open(self.pending_daemons, 'r') as fh:
            for line in fh.readlines():
                line = line.rstrip()

                if not line or line.startswith('#'):
                    continue

                if '=' in line:
                    (daemon, status) = line.split('=')
                    daemon = daemon.strip()
                    status = status.strip()
                    daemons[daemon] = status

        # zebra must always be enabled
        if daemons.get('zebra') != 'yes':
            daemons['zebra'] = 'yes'
            rewrite_daemons = True

        # target_context is None for commands that go in the global
        # context (static routes, debugs, etc)
        if add_line:

            # If this is multicast related command then pimd must be enabled
            if ('igmp' in add_line or 'pim' in add_line or 'multicast' in add_line) and os.path.exists(usr_lib_quagga_pimd):
                if daemons.get('pimd') != 'yes':
                    daemons['pimd'] = 'yes'
                    rewrite_daemons = True

            if target_context:

                # Enable bgpd if needed
                if target_context.startswith('router bgp'):
                    if daemons.get('bgpd') != 'yes':
                        daemons['bgpd'] = 'yes'
                        rewrite_daemons = True

                # Enable ospfd if needed
                elif target_context == 'router ospf':
                    if daemons.get('ospfd') != 'yes':
                        daemons['ospfd'] = 'yes'
                        rewrite_daemons = True

                elif target_context == 'router ospf6':
                    if daemons.get('ospf6d') != 'yes':
                        daemons['ospf6d'] = 'yes'
                        rewrite_daemons = True

                elif target_context.startswith('interface'):
                    if 'ip ospf' in add_line:
                        if daemons.get('ospfd') != 'yes':
                            daemons['ospfd'] = 'yes'
                            rewrite_daemons = True

                    elif 'ipv6 ospf6' in add_line:
                        if daemons.get('ospf6d') != 'yes':
                            daemons['ospf6d'] = 'yes'
                            rewrite_daemons = True

            else:
                if 'debug ospf ' in add_line:
                    if daemons.get('ospfd') != 'yes':
                        daemons['ospfd'] = 'yes'
                        rewrite_daemons = True

                elif 'debug ospf6 ' in add_line:
                    if daemons.get('ospf6d') != 'yes':
                        daemons['ospf6d'] = 'yes'
                        rewrite_daemons = True

                elif 'debug bgp ' in add_line:
                    if daemons.get('bgpd') != 'yes':
                        daemons['bgpd'] = 'yes'
                        rewrite_daemons = True

        if rewrite_daemons:
            with open(self.pending_daemons, 'w') as fh:
                for daemon in daemons.keys():
                    fh.write("%s=%s\n" % (daemon, daemons[daemon]))

        # Stage config changes to pending_conf
        # r+ opens the file for reading and writing
        with open(self.pending_conf, 'r+') as fh:

            # Read in all lines and the reset the fh pointer back to
            # the start of the file
            lines = fh.readlines()
            fh.seek(0)

            state = 'LIMBO'
            subcontext_lines = []

            # Now write all of the lines back to the file but add the add_line
            # (if there is one) at the end of its context/subcontext.
            for line in lines:
                line = line.rstrip()

                if state == 'LIMBO':
                    # If we are adding lines at the global level (static routes, debugs, etc)
                    # then target_context will be None
                    if target_context:

                        if line == target_context:
                            state = 'IN_CONTEXT'

                elif state == 'IN_CONTEXT':
                    if target_subcontext and line.strip() == target_subcontext:
                        state = 'IN_SUBCONTEXT'

                    elif line.strip() == 'end':
                        # We just finished writing all of the lines for our context, if
                        # add_line wasn't already in there then write it too
                        if add_line:

                            # If our add_line is CONTEXT there isn't anything
                            # to do because the context already exists
                            if add_line != 'CONTEXT':
                                if add_line in subcontext_lines:
                                    print "'%s' configuration already has '%s'" %\
                                        (target_context, add_line)
                                else:

                                    # Create the subcontext
                                    if target_subcontext:
                                        fh.write(' ' + target_subcontext + '\n')
                                        fh.write('  ' + add_line + '\n')
                                        fh.write(' exit-address-family\n')
                                    else:
                                        fh.write(' ' + add_line + '\n')
                            add_line = None

                        # If we are deleting the entire context we must also delete the 'end'
                        # Note that the 'interface' check here is because in quagga we cannot
                        # do a "no" on an entire interface. Tracking this via CM-11563
                        elif del_line == 'CONTEXT' and not target_context.startswith('interface '):
                            state = 'LIMBO'
                            continue

                        state = 'LIMBO'

                elif state == 'IN_SUBCONTEXT':
                    if line.strip() == 'exit-address-family' or line.strip() == 'end':

                        # We just finished writing all of the lines for our subcontext, if
                        # add_line wasn't already in there then write it too
                        if add_line and add_line != 'CONTEXT':
                            if add_line in subcontext_lines:
                                print "'%s %s' configuration already has '%s'" %\
                                    (target_context, target_subcontext, add_line)
                            else:
                                fh.write('  ' + add_line + '\n')
                        add_line = None
                        state = 'IN_CONTEXT'

                    elif line.strip() == 'end':
                        state = 'LIMBO'

                log.debug("%s: %s" % (state, line))

                if (not target_context or
                    (not target_subcontext and state == 'IN_CONTEXT') or
                    (target_subcontext and state == 'IN_SUBCONTEXT')):

                    # We are deleting the entire context so do not write any of the lines
                    # Note that the 'interface' check here is because in quagga we cannot
                    # do a "no" on an entire interface. Tracking this via CM-11563
                    if del_line == 'CONTEXT' and not line.startswith('interface '):
                        pass

                    # Do we write the current line?
                    elif exact_match and line.strip() == del_line:
                        found_del_line = True

                    elif not exact_match and line.strip().startswith(del_line):
                        found_del_line = True

                    else:
                        # If the user is adding a line to the global context
                        # do not add the same line multiple times
                        if not target_context and line == add_line:
                            print "configuration already has '%s'" % add_line
                        else:
                            fh.write(line + '\n')

                    subcontext_lines.append(line.strip())
                else:
                    # If we are deleting the entire ospf context then we also
                    # need to delete the ospf lines under the interfaces
                    if del_line == 'CONTEXT' and target_context == 'router ospf' and line.lstrip().startswith('ip ospf '):
                        pass
                    elif del_line == 'CONTEXT' and target_context == 'router ospf6' and line.lstrip().startswith('ipv6 ospf6 '):
                        pass
                    else:
                        fh.write(line + '\n')

            if add_line:
                log.debug("STATE %s, ADD_LINE: %s" % (state, add_line))

                # Note that CONTEXT is a special case.  We hit this if
                # the user is trying to add a context such as "router bgp 100"
                if add_line == 'CONTEXT':
                    if state not in ('IN_CONTEXT', 'IN_SUBCONTEXT'):
                        fh.write('%s\n' % target_context)

                        if target_subcontext:
                            fh.write(' %s\n' % target_subcontext)
                            fh.write(' exit-address-family\n')
                        fh.write('end\n')

                # The context (or sub-context) that we were looking wasn't in the config
                else:

                    # Enter the context...if it is None that means global context
                    if target_context:
                        fh.write('%s\n' % target_context)

                        # We are in the context, enter the sub-context if there is
                        # one. Indent accordingly
                        if target_subcontext:
                            fh.write(' %s\n' % target_subcontext)
                            fh.write('  %s\n' % add_line)
                            fh.write(' exit-address-family\n')
                        else:
                            fh.write(' %s\n' % add_line)
                        fh.write('end\n')
                    else:
                        fh.write('%s\n' % add_line)
                        fh.write('end\n')

            # If we deleted lines we need to truncate the rest of the fh so that we
            # do not end up with the lines at the end printed twice
            fh.truncate()

        if not found_del_line and del_line != 'CONTEXT':
            if target_subcontext:
                print "'%s %s' configuration does not have '%s'" % (target_context, target_subcontext, del_line)
            elif target_context:
                print "'%s' configuration does not have '%s'" % (target_context, del_line)
            else:
                print "configuration does not have '%s'" % del_line

    def bgp_neighbor_sanity(self, context, subcontext, cmd):
        # If we are adding a keyword for a bgp neighbor there are a few sanity
        # checks that we must make
        # - Is the peer defined?
        # - If the peer's name matches an interface is the peer defined as an "interface" peer?
        cmd = cmd.split()

        if context and context.startswith('router bgp') and cmd[0].startswith('neighbor'):
            neighbor = cmd[1]
            keyword = cmd[2]
            interfaces = [x for x in listdir('/sys/class/net/') if x != 'bonding_masters']

            # interface peer
            if neighbor in interfaces:
                if keyword != 'interface' and not self.neighbor_interface(neighbor):
                    self.edit_pending_file(context, None,
                                           'neighbor %s interface' % neighbor,
                                           None, True)

            # ipv4/ipv6 peer
            else:
                if keyword not in ('remote-as', 'peer-group') and not self.neighbor_remote_as(neighbor):
                    self.edit_pending_file(context, None,
                                           'neighbor %s remote-as external' % neighbor,
                                           None, True)

    def add(self, context, subcontext, cmd):
        self.bgp_neighbor_sanity(context, subcontext, cmd)
        self.edit_pending_file(context, subcontext, cmd, None, True)

    def delete(self, context, subcontext, cmd, exact_match):
        self.edit_pending_file(context, subcontext, None, cmd, exact_match)

    def abort_pending(self):
        for filename in (self.baseline_conf, self.pending_conf, self.pending_baseline_conf,
                         self.baseline_daemons, self.pending_daemons):
            if os.path.isfile(filename):
                os.unlink(filename)

    def commit_pending(self, verbose):
        """
        Make changes persistent by running quagga-reload.py
        """
        files_modified = []
        restart_quagga = False

        if (not os.path.isfile(self.pending_baseline_conf) or
            not os.path.isfile(self.pending_conf) or
            not self.init_guts()):
            return files_modified

        # If there is a delta, /usr/bin/diff will raise CalledProcessError
        try:
            check_output(['/usr/bin/diff', self.baseline_daemons, self.pending_daemons])

            # If we are here, there wasn't a change to /etc/quagga/daemons
            try:
                # If quagga is enabled but not active then we need to start quagga
                check_output(['/bin/systemctl', 'is-enabled', 'quagga'])

                try:
                    check_output(['/bin/systemctl', 'is-active', 'quagga'])

                # is-active Exception
                except CalledProcessError as e:
                    restart_quagga = True

            # is-enabled Exception
            except CalledProcessError as e:
                pass

        # /usr/bin/diff Exception
        except CalledProcessError as e:
            copyfile(self.pending_daemons, etc_quagga_daemons)
            chown_chgroup(etc_quagga_daemons, 'quagga', 'quagga')

            # A new daemon was enabled in /etc/quagga/daemons so we must
            # restart quagga so that the new daemon is started. Post CM-11625
            # we can do a start here which will be much less disruptive.
            restart_quagga = True

        if restart_quagga:

            try:
                output = check_output(['/bin/systemctl', 'reset-failed', 'quagga'])
                if verbose:
                    print output

                output = check_output(['/bin/systemctl', 'restart', 'quagga'])
                if verbose:
                    print output

            except Exception:
                print "ERROR: Failed to restart quagga"

            files_modified.append(etc_quagga_daemons)

        os.unlink(self.baseline_daemons)
        os.unlink(self.pending_daemons)

        # If there is a delta, /usr/bin/diff will raise CalledProcessError
        # If there is not a delta to apply do not bother calling quagga-reload.py
        try:
            check_output(['/usr/bin/diff', self.pending_baseline_conf, self.pending_conf])
            os.unlink(self.baseline_conf)
            os.unlink(self.pending_baseline_conf)
            os.unlink(self.pending_conf)
            return files_modified
        except CalledProcessError as e:
            pass

        try:
            # The /var/run/nclu/quagga/ directly is locked down (we want to keep it this way)
            # but /var/run/nclu/ is open for read so cp pending_conf to /var/run/nclu/ so it
            # is in a place where quagga-reload.py can read it
            tmp_pending_conf = '/var/run/nclu/quagga_pending.conf'
            copyfile(self.pending_conf, tmp_pending_conf)

            if self.show_linux_command:
                print "Updating %s and applying changes via quagga-reload.py" % etc_quagga_quagga_conf

            output = check_output(['/usr/lib/quagga/quagga-reload.py',
                                   '--stdout',
                                   '--reload',
                                   '/var/run/nclu/quagga_pending.conf'], stderr=STDOUT)
            os.unlink(tmp_pending_conf)

            if verbose:
                print output

            os.unlink(self.baseline_conf)
            os.unlink(self.pending_baseline_conf)
            os.unlink(self.pending_conf)
            files_modified.append(etc_quagga_quagga_conf)
        except CalledProcessError as e:
            log.exception(e)
            log.error(e.output)
            print "\nERROR: quagga-reload.py failed, is quagga running? Check 'systemctl status quagga'"
            print "\n%s\n%s\n%s\n" % ('=' * 80, e.output, '=' * 80)

            # quagga-reload failed to apply all staged changes so we must use
            # quagga-reload to undo any changes that did get applied. We do this
            # by reloading the existing /etc/quagga/Quagga.conf
            try:
                check_output(['/usr/lib/quagga/quagga-reload.py',
                              '--stdout',
                              '--reload',
                              etc_quagga_quagga_conf], stderr=STDOUT)
            except CalledProcessError as e:
                log.exception(e)
                log.error(e.output)
                print "\nERROR: quagga-reload.py failed to restore %s" % etc_quagga_quagga_conf
                print "\n%s\n%s\n%s\n" % ('=' * 80, e.output, '=' * 80)

        return files_modified

    def show_pending(self, print_errors=True):

        if not self.init_guts(print_missing_file_errors=False):
            return False

        diff_pending = False

        if print_diff(self.baseline_daemons, self.pending_daemons, self.color_diffs):
            diff_pending = True
            print """\033[91m\n
********************************************************************
********************************************************************
NOTE: 'net commit' will run 'systemctl restart quagga' to apply the
change to %s
********************************************************************
********************************************************************
\033[0m
""" % etc_quagga_daemons


        if print_diff(self.pending_baseline_conf, self.pending_conf, self.color_diffs):
            diff_pending = True

        return diff_pending

    def show_config_files(self, user_may_edit):

        if os.path.exists(etc_quagga_daemons):
            print "%s" % etc_quagga_daemons
            print "==================="

            with open(etc_quagga_daemons, 'r') as fh:
                print ''.join(fh.readlines())
            print ''

        if os.path.exists(etc_quagga_quagga_conf):
            print "%s" % etc_quagga_quagga_conf
            print "======================="

            # If the user has edit privs then do not bother hiding passwords
            if user_may_edit:
                with open(etc_quagga_quagga_conf, 'r') as fh:
                    print ''.join(fh.readlines())

            # If they can only do "show" commands then replace any passwords with PASSWORD_REMOVED
            else:
                output = []
                with open(etc_quagga_quagga_conf, 'r') as fh:
                    for line in fh.readlines():

                        if 'authentication-key' in line:
                            re_ospf_auth_key = re.search('^(.* ospf authentication-key\s+)\S+(.*)$', line)
                            if re_ospf_auth_key:
                                line = re_ospf_auth_key.group(1) + 'PASSWORD_REMOVED' + re_ospf_auth_key.group(2) + '\n'
                                output.append(line)
                            else:
                                output.append(line)

                        elif 'password' in line and 'password 8 ' not in line:
                            re_password = re.search('^(.*password\s+)\S+(.*)$', line)

                            if re_password:
                                line = re_password.group(1) + 'PASSWORD_REMOVED' + re_password.group(2) + '\n'
                                output.append(line)
                            else:
                                output.append(line)
                        else:
                            output.append(line)
                print ''.join(output)
            print ''

    def show_config_commands(self, user_may_edit):
        interface_commands = []
        commands = []
        supported = []
        not_supported = []

        if os.path.exists(etc_quagga_quagga_conf):
            try:
                quagga_conf = quagga_reload.Config()
                quagga_conf.load_from_file(etc_quagga_quagga_conf)
            except quagga_reload.VtyshMarkException as e:
                msg = "Failed to load %s due to %s" % (etc_quagga_quagga_conf, str(e))
                log.error(msg)
                print msg
                return (supported, not_supported)

            for context in quagga_conf.contexts.itervalues():

                if len(context.keys) == 1:
                    key = context.keys[0]

                    if key.startswith('router bgp'):

                        if 'vrf' in key:
                            # router bgp 100 vrf RED
                            vrf = key.split()[4]
                        else:
                            vrf = None

                        printf_alt = '%s\\n' % key
                        commands.append((printf_alt, 'net add %s' % key))

                        for line in context.lines:
                            printf_alt = '%s\\n  %s\\n' % (key, line)

                            if vrf:
                                cmd = 'net add bgp vrf %s %s' % (vrf, line)
                            else:
                                cmd = 'net add bgp %s' % line

                            cmd = cmd.replace(' bgp bgp ', ' bgp ')
                            commands.append((printf_alt, cmd))

                    elif key.startswith('router ospf6'):
                        printf_alt = '%s\\n' % key
                        commands.append((printf_alt, 'net add %s' % key))

                        for line in context.lines:
                            printf_alt = '%s\\n  %s\\n' % (key, line)
                            commands.append((printf_alt, 'net add ospf6 %s' % line))

                    elif key.startswith('router ospf'):
                        printf_alt = '%s\\n' % key
                        commands.append((printf_alt, 'net add %s' % key))

                        for line in context.lines:
                            printf_alt = '%s\\n  %s\\n' % (key, line)
                            cmd = 'net add ospf %s' % line
                            cmd = cmd.replace(' ospf ospf ', ' ospf ')
                            commands.append((printf_alt, cmd))

                    elif context.lines:
                        if key.startswith('interface'):
                            for line in context.lines:
                                interface_commands.append('net add %s %s' % (key, line))
                        else:
                            for line in context.lines:
                                printf_alt = '%s\\n  %s\\n' % (key, line)
                                commands.append((printf_alt, 'net add %s %s' % (key, line)))
                    else:
                        printf_alt = '%s\\n' % key
                        cmd = 'net add %s' % key
                        commands.append((printf_alt, 'net add %s' % key))

                elif len(context.keys) == 2:
                    key = context.keys[0]
                    subkey = context.keys[1]

                    if key.startswith('router bgp'):

                        if 'vrf' in key:
                            # router bgp 100 vrf RED
                            vrf = key.split()[4]
                        else:
                            vrf = None

                        if subkey == 'address-family ipv4 unicast':
                            for line in context.lines:
                                printf_alt = '%s\\n  %s\\n  %s\\n' % (key, subkey, line)

                                if vrf:
                                    commands.append((printf_alt, 'net add bgp vrf %s ipv4 unicast %s' % (vrf, line)))
                                else:
                                    commands.append((printf_alt, 'net add bgp ipv4 unicast %s' % line))

                        elif subkey == 'address-family ipv6 unicast':
                            for line in context.lines:
                                printf_alt = '%s\\n  %s\\n  %s\\n' % (key, subkey, line)

                                if vrf:
                                    commands.append((printf_alt, 'net add bgp vrf %s ipv6 unicast %s' % (vrf, line)))
                                else:
                                    commands.append((printf_alt, 'net add bgp ipv6 unicast %s' % line))

                        else:
                            log.info("quagga show_config_commands() unsupported BGP subkey %s" % subkey)
                    else:
                        log.info("quagga show_config_commands() unsupported key %s" % pformat(context.keys))
                else:
                    log.info("quagga show_config_commands() unsupported key %s" % pformat(context.keys))


        # Create a temporary NetworkDocopt object so we can see
        # which 'commands' are supported. We need to replace the
        # 'quagga' in our __doc__ with 'net'
        cli = NetworkDocopt(__doc__.replace('    quagga ', '    net '))

        interface_commands = globify_interface_commands(interface_commands)
        for cmd in interface_commands:

            # We do not want the user to see the "Command not found"
            # error if this is an unsupported command so use Capturing
            # but do nothing with the output.
            with Capturing() as _:
                cli.run(cmd.split())

            if cli.match:
                supported.append(cmd)
            else:
                cmd = cmd.split()
                ifname_glob = cmd[3]
                cmd = ' '.join(cmd[4:])

                for ifname in ifname_expand_glob(ifname_glob):
                    echo_cmd = "printf 'interface %s\\n%s\\n' >> %s" % (ifname, cmd, etc_quagga_quagga_conf)
                    not_supported.append(echo_cmd)
            cli.re_init()

        for (printf_alt, cmd) in commands:
            if cmd == '#':
                supported.append(cmd)
                continue

            # We do not want the user to see the "Command not found"
            # error if this is an unsupported command so use Capturing
            # but do nothing with the output.
            with Capturing() as _:
                cli.run(cmd.split())

            if cli.match:
                supported.append(cmd)
            else:
                echo_cmd = "printf '%s' >> %s" % (printf_alt, etc_quagga_quagga_conf)
                not_supported.append(echo_cmd)
            cli.re_init()

        return (supported, not_supported)

    def show_config_summary(self, user_may_edit, summary):

        # Nothing to do so return
        if not os.path.exists(etc_quagga_quagga_conf):
            return

        try:
            quagga_conf = quagga_reload.Config()
            quagga_conf.load_from_file(etc_quagga_quagga_conf)
        except NameError:
            # This is a temp work-around if we are running on a system where
            # quagga-reload.py is missing the "logger = logging.getLogger(__name__)" line
            lines = []
            with open('/usr/lib/quagga/quagga-reload.py', 'r') as fh:
                lines = fh.readlines()

            with open('/usr/lib/quagga/quagga-reload.py', 'w') as fh:
                for line in lines:
                    if line.startswith('class Context'):
                        fh.write('logger = logging.getLogger(__name__)\n\n')
                    fh.write(line)

            reload(quagga_reload)
            quagga_conf = quagga_reload.Config()
            quagga_conf.load_from_file(etc_quagga_quagga_conf)

        except quagga_reload.VtyshMarkException as e:
            msg = "Failed to load %s due to %s" % (etc_quagga_quagga_conf, str(e))
            log.error(msg)
            print msg
            return

        for context in quagga_conf.contexts.itervalues():

            if len(context.keys) == 1:
                key = context.keys[0]

                if key.startswith('interface '):
                    key = key.replace('interface ', 'INTERFACE:')

                if key not in summary:
                    summary[key] = OrderedDict()

                if context.lines:
                    if 'COMMAND_LIST' not in summary[key]:
                        summary[key]['COMMAND_LIST'] = []
                    summary[key]['COMMAND_LIST'].extend(context.lines)

                    if key.startswith('INTERFACE:'):
                        summary[key]['COMMAND_LIST'] = sort_for_humans(summary[key]['COMMAND_LIST'])

            elif len(context.keys) == 2:
                key = context.keys[0]
                subkey = context.keys[1]

                if key not in summary:
                    summary[key] = OrderedDict()

                if subkey not in summary[key]:
                    summary[key][subkey] = []

                summary[key][subkey] = context.lines
            else:
                log.info("quagga show_config_commands() unsupported key %s" % pformat(context.keys))


    def ip_route_show(self, vtysh_args, zebra_output):

        def get_routing_entry_for():
            for line in zebra_output.splitlines():
                if line.startswith('Routing entry for'):
                    return line.strip().split()[3]
            return None

        # show ip route x.x.x.x/24
        # ip route show x.x.x.x/24
        #
        # show ip route x.x.x.x
        # ip route show
        if vtysh_args.startswith('show ip route'):
            re_prefix_prefixlen = re_search('^show ip route (\d+\.\d+\.\d+\.\d+\/\d+)$', vtysh_args)
            re_prefix = re_search('^show ip route (\d+\.\d+\.\d+\.\d+)$', vtysh_args)
            use_ipv6 = False

        elif vtysh_args.startswith('show ipv6 route'):
            re_prefix_prefixlen = re_search('^show ipv6 route ([0-9a-fA-F]\S+\/\d+)$', vtysh_args)
            re_prefix = re_search('^show ipv6 route ([0-9a-fA-F]\S+:\S+)$', vtysh_args)
            use_ipv6 = True

        else:
            re_prefix_prefixlen = None
            re_prefix = None
            use_ipv6 = False

        if re_prefix_prefixlen:
            ip = re_prefix_prefixlen.group(1)

        elif re_prefix:
            ip = re_prefix.group(1)
            ip = get_routing_entry_for()
        else:
            ip = None

        if ip:
            header = "FIB entry for %s" % vtysh_args.split()[-1]
            print "%s\n%s" % (header, '=' * len(header))

            if self.show_linux_command:
                if use_ipv6:
                    print "exec: '/sbin/ip -6 route show %s'" % ip
                else:
                    print "exec: '/sbin/ip route show %s'" % ip

            try:
                if use_ipv6:
                    print check_output(['/sbin/ip', '-6', 'route', 'show', ip])
                else:
                    print check_output(['/sbin/ip', 'route', 'show', ip])
            except Exception:
                pass

    def del_interfaces(self, ifname):
        ifnames = []

        if ifname_is_glob(str(ifname)):
            for x in ifname_expand_glob(str(ifname)):
                ifnames.append(x)

        elif ifname_is_valid(str(ifname)) and not ifname_is_globish(str(ifname)):
            ifnames = [ifname, ]

        for ifname in ifnames:
            context = 'interface %s' % ifname
            subcontext = None
            cmd = 'CONTEXT'
            self.delete(context, subcontext, cmd, True)

    def exec_vtysh_show_or_clear(self, args, vtysh_args):
        # In quagga, bgp allows the user to do "show ip bgp" or "show bgp"
        # but ospf only allows "show ip ospf".  This is a bit awkward for
        # the user if they are working with both so we made "show ospf" but
        # we need to change that to "show ip ospf"
        vtysh_args = vtysh_args.replace('show ospf6', 'show ipv6 ospf6')
        vtysh_args = vtysh_args.replace('show ospf', 'show ip ospf')
        vtysh_args = vtysh_args.replace('show igmp', 'show ip igmp')
        vtysh_args = vtysh_args.replace('show mroute', 'show ip mroute')
        vtysh_args = vtysh_args.replace('show pim', 'show ip pim')
        vtysh_args = vtysh_args.replace('show msdp', 'show ip msdp')

        vtysh_args = vtysh_args.replace('clear ospf', 'clear ip ospf')
        vtysh_args = vtysh_args.replace('clear bgp', 'clear ip bgp')
        vtysh_args_list = []

        if vtysh_args.startswith('show bgp'):
            # Have this dump both IPv4 and IPv6 tables
            if vtysh_args == 'show bgp':
                vtysh_args_list.append('show bgp ipv4 unicast')
                vtysh_args_list.append('show bgp ipv6 unicast')

            # Have this dump both IPv4 and IPv6 summary
            elif vtysh_args == 'show bgp summary':
                vtysh_args_list.append('show bgp ipv4 unicast summary')
                vtysh_args_list.append('show bgp ipv6 unicast summary')

            elif vtysh_args == 'show bgp update-groups':
                vtysh_args_list.append('show bgp ipv4 unicast update-groups')
                vtysh_args_list.append('show bgp ipv6 unicast update-groups')

            # Nothing special for these, just run as they are
            elif (vtysh_args.startswith('show bgp ipv4 unicast') or
                  vtysh_args.startswith('show bgp ipv6 unicast') or
                  vtysh_args.startswith('show bgp neighbor')):
                vtysh_args_list.append(vtysh_args)

            # If we are here then we are running this command:
            # quagga show bgp (<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>) [bestpath|multipath] [json]
            #
            # So we must look to see if we are dealing with an IPv4 prefix or an IPv6 prefix
            # and add 'ipv4 unicast' or 'ipv6 unicast' to the command
            else:

                # For now vrf is ipv4 only but you can only do
                if args.get('vrf'):
                    vtysh_args = vtysh_args.replace('show bgp', 'show ip bgp')

                elif args.get('<ipv4>') or args.get('<ipv4/prefixlen>'):
                    vtysh_args = vtysh_args.replace('show bgp', 'show bgp ipv4 unicast')

                elif args.get('<ipv6>') or args.get('<ipv6/prefixlen>'):
                    vtysh_args = vtysh_args.replace('show bgp', 'show bgp ipv6 unicast')

                vtysh_args_list.append(vtysh_args)

        # quagga show route [vrf <text>] [<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>|bgp|connected|ospf|static|summary]
        # quagga show route [vrf <text>] (<ipv4/prefixlen>|<ipv6/prefixlen>) longer-prefixes
        elif vtysh_args.startswith('show route') and not vtysh_args.startswith('show route-map'):

            if vtysh_args == 'show route':
                vtysh_args_list.append('show ip route')
                vtysh_args_list.append('show ipv6 route')

            else:
                if args.get('<ipv6>') or args.get('<ipv6/prefixlen>') or args.get('ospf6') or args.get('ripng') or args.get('ipv6'):
                    vtysh_args = vtysh_args.replace('show route', 'show ipv6 route')

                else:
                    vtysh_args = vtysh_args.replace('show route', 'show ip route')

                # When ip_route_show() is called we will print the kernel entry for this route
                # so print a little header here so the user can tell which is which
                header = "RIB entry for %s" % vtysh_args.split()[-1]
                print "%s\n%s" % (header, '=' * len(header))
                vtysh_args_list.append(vtysh_args)

        # Change "clear ip bgp all" to "clear ip bgp *"
        elif 'clear' in vtysh_args and 'bgp' in vtysh_args and 'all' in vtysh_args:
            vtysh_args_list.append(vtysh_args.replace(' all', ' *'))

        else:
            vtysh_args_list.append(vtysh_args)

        print_header = True if len(vtysh_args_list) > 1 else False

        for vtysh_args in vtysh_args_list:
            cmd = ['/usr/bin/vtysh', '-c', vtysh_args]

            if self.show_linux_command and 'json' not in vtysh_args:
                print "exec 'vtysh -c \"%s\"'\n" % vtysh_args

            if print_header:
                print "\n%s\n%s" % (vtysh_args, '=' * len(vtysh_args))

            try:
                output = check_output(cmd)
            except Exception as e:
                log.error("'%s' returned '%s' with error:\n%s\n" % (' '.join(cmd), e.output.rstrip(), e))
                print e.output
                output = None

            if output:
                print output

        # For "show route" also show the kernel route
        if vtysh_args.startswith('show ip route') or vtysh_args.startswith('show ipv6 route'):
            self.ip_route_show(vtysh_args, output)


    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the self.args dictionary based on the command
        line args entered by the user
        """
        command_ok = True
        args = cli.args

        # If the user does "net show route" when quagga isn't running don't print
        # an error just return blank
        if args.get('show') or args.get('clear'):
            print_missing_file_errors = False
        else:
            print_missing_file_errors = True

        if not self.init_guts(print_missing_file_errors):
            return False

        bgp_bgp = ['always-compare-med',
                   'bestpath',
                   'client-to-client reflection',
                   'cluster-id',
                   'confederation peers',
                   'default ipv4-unicast',
                   'default local-preference',
                   'default show-hostname',
                   'default subgroup-pkt-queue-max',
                   'deterministic-med',
                   'disable-ebgp-connected-route-check',
                   'enforce-first-as',
                   'fast-external-failover',
                   'graceful-restart',
                   'listen',
                   'log-neighbor-changes',
                   'max-med',
                   'network import-check',
                   'route-map delay-timer',
                   'route-reflector allow-outbound-policy',
                   'router-id']
        ospf_ospf = ['router-id', ]
        ifnames = []
        bgppeers = []

        # If <interface> is a glob then populate ifnames with all of the
        # interface names produced by that glob else populate ifnames with <interface>
        if args.get('<interface>'):
            ifname = args.get('<interface>')

            if ifname_is_glob(str(ifname)):
                for x in ifname_expand_glob(str(ifname)):
                    ifnames.append(x)

            elif ifname_is_valid(str(ifname)) and not ifname_is_globish(str(ifname)):
                ifnames = [ifname, ]

            else:
                print "ERROR: a valid interface name or glob is needed"
                return False

        elif args.get('<bgppeer>'):
            bgppeer = str(args.get('<bgppeer>'))

            if ifname_is_glob(bgppeer):
                for x in ifname_expand_glob(bgppeer):
                    bgppeers.append(x)
            else:
                bgppeers.append(bgppeer)

        if args.get('show') or args.get('clear'):
            cmd = []

            for node in cli.match[1:]:
                if node.value:
                    if node.value.text == '<interface>':
                        cmd.append(node.value.text)
                    elif node.value.text.startswith('<'):
                        cmd.append(node.value.value)
                    else:
                        cmd.append(node.value.text)

                # Default afi to 'ipv4' and safi to 'unicast'
                else:
                    if str(node) == '[ipv4|ipv6]':
                        cmd.append('ipv4')
                    elif str(node) == '[unicast|multicast]':
                        cmd.append('unicast')
                    elif str(node) == '[ipv4 unicast|ipv6 unicast]':
                        cmd.append('ipv4')
                        cmd.append('unicast')

            vtysh_args = ' '.join(map(str, cmd))

            # Support show commands with globs of interfaces
            # net show ospf interface swp3-4
            if ifnames:
                for ifname in ifnames:
                    vtysh_args_for_ifname = vtysh_args.replace('<interface>', ifname)
                    self.exec_vtysh_show_or_clear(args, vtysh_args_for_ifname)
            elif bgppeers:
                for peer in bgppeers:
                    vtysh_args_for_bgppeer = vtysh_args.replace('<bgppeer>', peer)
                    self.exec_vtysh_show_or_clear(args, vtysh_args_for_bgppeer)
            else:
                self.exec_vtysh_show_or_clear(args, vtysh_args)

        elif args.get('add') or args.get('del'):
            contexts = []
            subcontext = None

            # (add|del) will be the node at #0
            # entry #1 will be the bgp, ospf, interface, etc node
            add_del_node = cli.match[1]
            context_node = cli.match[2]

            # Set node_cmd_start to 3 so we don't look at "(add|del) bgp" when
            # building the actual quagga cmd string
            node_cmd_start = 3

            # If the user entered "bgp autonomous-system 100"
            # We must translate that to "router bgp 100"
            if context_node.value.text == 'bgp' and cli.match[3].value:
                keyword = cli.match[3].value.text

                if keyword == 'autonomous-system':
                    cli.match[2].value.value = 'router'
                    cli.match[2].value.value_raw = 'router'
                    cli.match[3].value.value = 'bgp'
                    cli.match[3].value.value_raw = 'bgp'

                # net add bgp    vrf BLUE autonomous-system 77
                # net add router bgp 77   vrf               BLUE
                # 0   1   2      3   4    5                 6
                elif keyword == 'vrf':
                    if len(cli.match) >= 7:
                        if cli.match[5].value.text == 'autonomous-system':
                            vrf = cli.match[4].value.value
                            asn = cli.match[6].value.value
                            cli.match[2].value.value = 'router'
                            cli.match[2].value.value_raw = 'router'
                            cli.match[3].value.value = 'bgp'
                            cli.match[3].value.value_raw = 'bgp'
                            cli.match[4].value.value = asn
                            cli.match[4].value.value_raw = asn
                            cli.match[5].value.value = 'vrf'
                            cli.match[5].value.value_raw = 'vrf'
                            cli.match[6].value.value = vrf
                            cli.match[6].value.value_raw = vrf

            cmd_is_CONTEXT = False

            # Find the context and subcontext
            if context_node.value.value == 'router':
                protocol = cli.match[3].value.value

                if len(cli.match) > 6:
                    asn = cli.match[4].value.value
                    vrf = cli.match[6].value.value
                    contexts.append("router %s %d vrf %s" % (protocol, asn, vrf))

                elif len(cli.match) > 4:
                    asn = cli.match[4].value.value
                    contexts.append("router %s %d" % (protocol, asn))
                else:
                    contexts.append("router %s" % protocol)

                cmd_is_CONTEXT = True

            elif context_node.value.value == 'bgp':
                bgp_vrf = None
                afi_idx = 3
                safi_idx = 4

                if args.get('vrf'):
                    bgp_vrf = args.get('<text>')
                    node_cmd_start += 2
                    afi_idx += 2
                    safi_idx += 2

                bgp_asn = self.get_bgp_asn(bgp_vrf)

                if not bgp_asn:
                    print "ERROR: You must first enable BGP via 'net add bgp autonomous-system NUMBER'"
                    return False

                if bgp_vrf:
                    contexts.append('router bgp %d vrf %s' % (bgp_asn, bgp_vrf))
                else:
                    contexts.append('router bgp %d' % bgp_asn)

                # find the subcontext for bgp
                if str(cli.match[node_cmd_start]) == '[ipv4|ipv6]':
                    ipv4_ipv6_node = cli.match[afi_idx]
                    uni_multi_node = cli.match[safi_idx]

                    if ipv4_ipv6_node.value and ipv4_ipv6_node.value.text:
                        afi = ipv4_ipv6_node.value.text
                    else:
                        afi = 'ipv4'

                    if uni_multi_node.value and uni_multi_node.value.text:
                        safi = uni_multi_node.value.text
                    else:
                        safi = 'unicast'

                    subcontext = "address-family %s %s" % (afi, safi)
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == '[ipv4]' and str(cli.match[safi_idx]) == '[unicast]':
                    subcontext = "address-family ipv4 unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == '[ipv6]' and str(cli.match[safi_idx]) == '[unicast]':
                    subcontext = "address-family ipv6 unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == 'ipv4' and str(cli.match[safi_idx]) == 'unicast':
                    subcontext = "address-family ipv4 unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == 'ipv6' and str(cli.match[safi_idx]) == 'unicast':
                    subcontext = "address-family ipv6 unicast"
                    node_cmd_start += 2

                # The "ipv4 unicast" is optional (just to make v4 easier to configure)
                # for any AFI/SAFI specific command. If they left off the "ipv4 unicast"
                # add it for them.
                if subcontext is None:
                    for x in ('aggregate-address',
                              'network',
                              'redistribute',
                              'maximum-paths',
                              'activate',
                              'allowas-in',
                              'default-originate',
                              'filter-list',
                              'next-hop-self',
                              'prefix-list',
                              'remove-private-AS',
                              'route-map',
                              'route-reflector-client',
                              'send-community',
                              'soft-reconfiguration'):
                        if x in cli.argv_expanded:
                            subcontext = "address-family ipv4 unicast"
                            break

            elif context_node.value.text == 'ospf':
                contexts.append('router ospf')

            elif context_node.value.text == 'ospf6':
                contexts.append('router ospf6')

            elif context_node.value.text == 'interface':
                if ifnames:
                    for ifname in ifnames:
                        contexts.append('interface %s' % ifname)
                else:
                    contexts.append('interface %s' % cli.match[3].value.text)

                node_cmd_start += 1

            elif context_node.value.value == 'route-map':
                contexts.append(' '.join(map(str, cli.argv_expanded[2:6])))
                node_cmd_start += 3

                # 'net add route-map FOO permit 10' is 6 words
                if len(cli.argv_expanded) == 6:
                    cmd_is_CONTEXT = True

            # global config level
            else:
                node_cmd_start = 2
                contexts.append(None)

            if cmd_is_CONTEXT:
                cmd = 'CONTEXT'
            else:
                cmd = []
                for node in cli.match[node_cmd_start:]:
                    if node.value and node.value.text:
                        if node.value.text == '<interface>':
                            cmd.append('<interface>')
                        elif node.value.text == '<bgppeer>':
                            cmd.append('<bgppeer>')
                        else:
                            cmd.append(node.value.value)

                cmd = ' '.join(map(str, cmd))

            if contexts and contexts[0]:

                # There are about two dozen commands that are in the global bgp
                # context and start with the keyword 'bgp'. 'bgp router-id x.x.x.x'
                # is one example.  We did not want the parser to read:
                #
                #    quagga add bgp bgp router-id x.x.x.x
                #
                # Because the double bgp just looks/feels weird. Prepend the needed
                # bgp keyword for any of these two dozen commands.
                if contexts[0].startswith('router bgp'):
                    for bgp_bgp_cmd in bgp_bgp:
                        if cmd.startswith(bgp_bgp_cmd):
                            cmd = 'bgp ' + cmd
                            break

                    # 'bgp default ipv4-unicast' is on by default so if the user
                    # does a 'del' we need to convert that to an 'add' of 'no bgp default ipv4-unicast'
                    if cmd == 'bgp default ipv4-unicast':
                        cmd = 'no ' + cmd

                        if args.get('del'):
                            args['del'] = None
                            args['add'] = True
                        elif args.get('add'):
                            args['del'] = True
                            args['add'] = None

                # Same problem for some ospf commands
                elif contexts[0] == 'router ospf' or contexts[0].startswith('router ospf '):
                    for ospf_ospf_cmd in ospf_ospf:
                        if cmd.startswith(ospf_ospf_cmd):
                            cmd = 'ospf ' + cmd
                            break

                # Like 'bgp default ipv4-unicast', 'ipv6 nd suppress-ra' is another
                # one that is on by default so we have to jump through some hoops
                # to add the "no" line
                elif contexts[0].startswith('interface'):
                    if cmd == 'ipv6 nd suppress-ra':
                        cmd = 'no ' + cmd

                        if args.get('del'):
                            args['del'] = None
                            args['add'] = True
                        elif args.get('add'):
                            args['del'] = True
                            args['add'] = None

            # If the user is doing a del there are several options where we can
            # del based on the keyword and do not require the value.  For example
            # the user can delete the bgp router-id using either of the following:
            #
            #   net del bgp router-id 1.1.1.1
            #   net del bgp router-id
            #
            # exact_match determines if delete() looks for the line 'bgp router-id 1.1.1.1'
            # or just look for a line that begins with 'bgp router-id'
            if str(add_del_node) == '(add|del)' or str(add_del_node) == 'add':
                exact_match = True

            elif str(add_del_node) == 'del':
                last_node = cli.match[-1]

                # If the user specified the optional arg at the end then
                # exact_match must be True
                if not last_node.value:
                    exact_match = False
                else:
                    exact_match = True

            else:
                raise Exception("Invalid add_del_node %s" % add_del_node)

            if ifnames:
                # If the user entered a glob for interfaces those need to be
                # expanded in one of two places. The first scenario is where
                # the interface name is part of the context.  For example:
                #
                #    net add inter swp3-4 ip ospf network point-to-point
                #
                # needs to add this ospf command under contexts 'interface swp3
                # and 'interface swp4'. But if the user did
                #
                #    net add bgp neighbor swp3-4 interface peer-group FOO
                #
                # we need to expand this as separate commands within the 'router bgp' context.
                #
                # So check the len of contexts to figure out which way we
                # should expand the ifnames
                if len(contexts) > 1:
                    for context in contexts:
                        if args.get('add'):
                            self.add(context, subcontext, cmd)
                        elif args.get('del'):
                            self.delete(context, subcontext, cmd, exact_match)
                else:
                    context = contexts[0]

                    for ifname in ifnames:
                        cmd_with_ifname = cmd.replace('<interface>', ifname)

                        if args.get('add'):
                            self.add(context, subcontext, cmd_with_ifname)
                        elif args.get('del'):
                            self.delete(context, subcontext, cmd_with_ifname, exact_match)
            elif bgppeers:
                context = contexts[0]

                for peer in bgppeers:
                    cmd_with_peer = cmd.replace('<bgppeer>', peer)

                    if args.get('add'):
                        self.add(context, subcontext, cmd_with_peer)
                    elif args.get('del'):
                        self.delete(context, subcontext, cmd_with_peer, exact_match)
            else:
                context = contexts[0]

                if args.get('add'):
                    self.add(context, subcontext, cmd)
                elif args.get('del'):
                    self.delete(context, subcontext, cmd, exact_match)

        else:
            print "ERROR: no commands matched"
            command_ok = False

        return command_ok


def get_config_lines(target):
    final = []

    for filename in ('/var/run/nclu/quagga/pending.conf', etc_quagga_quagga_conf):
        if os.path.exists(filename):
            with open(filename, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith(target):
                        final.append(line)

    return final


def get_aspath_access_list():
    final = []

    for line in get_config_lines('ip as-path access-list'):
        # ip as-path access-list LARRY permit 1234
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_bgp_peergroups(args={}):
    final = []
    vrf = None

    if args.get('vrf'):
        vrf = "vrf %s" % args.get('<text>')

    for filename in ('/var/run/nclu/quagga/pending.conf', etc_quagga_quagga_conf):
        if os.path.exists(filename):
            with open(filename, 'r') as fh:
                state = None

                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith('router bgp'):
                        if vrf:
                            if line.endswith(vrf):
                                state = 'ROUTER_BGP'
                            else:
                                state = None
                        else:
                            state = 'ROUTER_BGP'

                    if line.startswith('neighbor ') and line.endswith('peer-group'):
                        if state == 'ROUTER_BGP':
                            final.append(line.split()[1])

    final = sort_for_humans(list(set(final)))
    return final


def get_bgppeers(args={}):
    final = []
    vrf = None

    if args.get('vrf'):
        vrf = "vrf %s" % args.get('<text>')

    for filename in ('/var/run/nclu/quagga/pending.conf', etc_quagga_quagga_conf):
        if os.path.exists(filename):
            with open(filename, 'r') as fh:
                state = None

                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith('router bgp'):
                        if vrf:
                            if line.endswith(vrf):
                                state = 'ROUTER_BGP'
                            else:
                                state = None
                        else:
                            state = 'ROUTER_BGP'

                    if line.startswith('neighbor '):
                        if state == 'ROUTER_BGP':
                            final.append(line.split()[1])

    final = sort_for_humans(list(set(final)))
    return final


def get_community_list():
    final = []

    for line in get_config_lines('ip community-list'):
        # ip community-list standard MARGE permit 100:100
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_extcommunity_list():
    final = []

    for line in get_config_lines('ip extcommunity-list'):
        # ip community-list standard MARGE permit 100:100
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_prefix_list_v4():
    final = []

    for line in get_config_lines('ip prefix-list'):
        # ip prefix-list BOB seq 5 permit any
        final.append(line.split()[2])

    final = sort_for_humans(list(set(final)))
    return final


def get_prefix_list_v6():
    final = []

    for line in get_config_lines('ipv6 prefix-list'):
        # ipv6 prefix-list JASON seq 5 permit any
        final.append(line.split()[2])

    final = sort_for_humans(list(set(final)))
    return final


def get_route_map():
    final = []

    for line in get_config_lines('route-map'):
        # route-map FOO permit 10
        final.append(line.split()[1])

    final = sorted(list(set(final)))
    return final


def match_aspath_access_list(argv_word, use_fuzzy, tab_completing):
    """
    <as-path-access-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_aspath_access_list():
            return True

        if tab_completing:
            for acl in get_aspath_access_list():
                if acl.startswith(argv_word):
                    return True
    return False


def match_bgppeer(argv_word, use_fuzzy, tab_completing):
    """
    <bgppeer>

    argv_word could be an IPv4 address, IPv6 address, ifname or glob of ifnames
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        potential_peers = []

        if ifname_is_glob(str(argv_word)):
            for x in ifname_expand_glob(str(argv_word)):
                potential_peers.append(x)
        else:
            potential_peers.append(argv_word)

        all_are_peers = True
        bgp_peers = get_bgppeers()

        for peer in potential_peers:
            if peer not in bgp_peers:
                all_are_peers = False

        if all_are_peers:
            return True

        if tab_completing and len(potential_peers) == 1:
            for peer in bgp_peers:
                if peer.startswith(argv_word):
                    return True
    return False


def match_community(argv_word, use_fuzzy, tab_completing):
    """
    <community>
    There is no fuzzy matching for communities
    """
    if re.search('^\d+:\d+$', argv_word):
        return True
    return False


def match_community_list(argv_word, use_fuzzy, tab_completing):
    """
    <community-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_community_list():
            return True

        if tab_completing:
            for commlist in get_community_list():
                if commlist.startswith(argv_word):
                    return True
    return False


def match_extcommunity(argv_word, use_fuzzy, tab_completing):
    """
    There is no fuzzy matching for extended communities
    ASN:nn_or_IP-address:nn
    """
    if re.search('^\d+:\d+$', argv_word):
        return True

    if re.search('^\d+\.\d+\.\d+\.\d+:\d+$', argv_word):
        return True

    return False


def match_extcommunity_list(argv_word, use_fuzzy, tab_completing):
    """
    <extcommunity-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_extcommunity_list():
            return True

        if tab_completing:
            for extcommlist in get_extcommunity_list():
                if extcommlist.startswith(argv_word):
                    return True
    return False


def match_route_map(argv_word, use_fuzzy, tab_completing):
    """
    There is no fuzzy matching for route-maps
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_route_map():
            return True

        if tab_completing:
            for routemap in get_route_map():
                if routemap.startswith(argv_word):
                    return True

    return False


def match_peer_group(argv_word, use_fuzzy, tab_completing):
    if ' ' not in argv_word:
        return True
    return False


def match_prefix_list_v4(argv_word, use_fuzzy, tab_completing):
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_prefix_list_v4():
            return True

        if tab_completing:
            for prefix_list in get_prefix_list_v4():
                if prefix_list.startswith(argv_word):
                    return True
    return False


def match_prefix_list_v6(argv_word, use_fuzzy, tab_completing):
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_prefix_list_v6():
            return True

        if tab_completing:
            for prefix_list in get_prefix_list_v6():
                if prefix_list.startswith(argv_word):
                    return True
    return False


def tab_aspath_access_list(args, ended_with_space, last_argv):
    final = []
    for x in get_aspath_access_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'as-path access-list'))

    if not final:
        final.append(('<as-path-access-list>', 'as-path access-list'))

    return final


def tab_bgppeer(args, ended_with_space, last_argv):
    final = []
    for x in get_bgppeers(args):
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP neighbor or peer-group'))

    if not final:
        final.append(('<bgppeer>', 'BGP neighbor or peer-group'))

    return final


def tab_community_list(args, ended_with_space, last_argv):
    final = []
    for x in get_community_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP community-list'))

    if not final:
        final.append(('<community-list>', 'BGP community-list'))

    return final


def tab_extcommunity_list(args, ended_with_space, last_argv):
    final = []
    for x in get_extcommunity_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP extcommunity-list'))

    if not final:
        final.append(('<extcommunity-list>', 'BGP extcommunity-list'))

    return final


def tab_peergroup(args, ended_with_space, last_argv):
    final = []
    for x in get_bgp_peergroups(args):
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP peer-group'))

    if not final:
        final.append(('<peergroup>', 'BGP peer-group'))

    return final


def tab_prefix_list_v4(args, ended_with_space, last_argv):
    final = []
    for x in get_prefix_list_v4():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'prefix-list'))

    if not final:
        final.append(('<prefix-list>', 'prefix-list'))

    return final


def tab_prefix_list_v6(args, ended_with_space, last_argv):
    final = []
    for x in get_prefix_list_v6():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'prefix-list'))

    if not final:
        final.append(('<prefix-list>', 'prefix-list'))

    return final


def tab_route_map(args, ended_with_space, last_argv):
    final = []

    # If ended_with_space is True the user did something like "net show route-map <TAB>"
    # so we need to list all route-maps.  If ended_with_space is False they did
    # "net show route-map B<TAB>" so we need to list only the route-maps that begin with 'B'
    for x in get_route_map():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'route-map'))

    if not final:
        final.append(('<route-map>', 'route-map'))

    return final
